﻿using System.Diagnostics;
using System.Reflection;
using System.Security.Cryptography;
using SharpFont;

namespace GraphicsManager.Objects.Core;

public class Font
{
	private List<Face> _Faces = new();
	internal static Library lib = new();
	public IReadOnlyList<Face> Fonts => _Faces.AsReadOnly();
	internal static Dictionary<string, Face> AllFileFonts = new();
	internal static Dictionary<byte[], Face> AllMemoryFonts = new();
	internal static List<string>? _SystemPre = null;
	private static List<Font> AllFonts = new();
	private static bool Backup = false;

	internal void AddSystemFontFace(string path)
	{
		Console.WriteLine("Added font: " + path);
		if (!AllFileFonts.ContainsKey(path)) AllFileFonts.Add(path, new(lib, path, 0));
		if (!_Faces.Contains(AllFileFonts[path]))
		{
			foreach (Font ft in AllFonts)
			{
				ft._Faces.Add(AllFileFonts[path]);
			}
		}
	}

	private Font()
	{
		Name = null!;
		Assembly = null!;
		Embeded = false;
		//TODO add more systemfont dection methods
		if (!Backup)
		{
			AllMemoryFonts.Add(SHA256.HashData(Tools.GetResourceBytes("GraphicsManager.Resources.Fonts.TektonPro-Regular.otf")), new Face(lib, Tools.GetResourceBytes("GraphicsManager.Resources.Fonts.TektonPro-Regular.otf"), 0));
			Backup = true;
		}
		if (_SystemPre is null)
		{
			_SystemPre = new();
			if (OperatingSystem.IsLinux())
			{
				try
				{
					Process proc = new()
					{
						StartInfo = new()
						{
							FileName = "/bin/bash",
							Arguments = "-c \"fc-list ':' file\"",
							RedirectStandardOutput = true,
						}
					};
					proc.Start();
					proc.WaitForExit();
					string[] files = proc.StandardOutput.ReadToEnd().Split($": {Environment.NewLine}");
					for (int i = 0; i < files.Length; i++)
					{
						_SystemPre.Add(files[i]);
					}
				}
				catch
				{
				}
			}
		}
	}
	public void SetEmbeddedFont(string Font, Assembly? Assembly = null)
	{
		_ = Font ?? throw new ArgumentNullException(nameof(Font));
		string Base = "GraphicsManager.Resources.Fonts.";
		if (Assembly is not null) Base = string.Empty;
		byte[] f = (Assembly is null
			? Tools.GetResourceBytes(Base + Font)
			: Tools.GetResourceBytes(Assembly!, $"{Base}{Font}"));
		byte[] hash = SHA256.HashData(f);
		if (AllMemoryFonts.ContainsKey(hash)) AllMemoryFonts.Add(hash, new(lib, f, 0));
		if (!_Faces.Contains(AllMemoryFonts[hash])) _Faces.Insert(0, AllMemoryFonts[hash]);
		else
		{
			_Faces.Remove(AllMemoryFonts[hash]);
			_Faces.Insert(0, AllMemoryFonts[hash]);
		}
		this.Assembly = Assembly;
		this.Embeded = true;
		this.Name = Font;
	}

	public static Font MakeEmbeddedFont(string Font, Assembly? Assembly = null)
	{
		_ = Font ?? throw new ArgumentNullException(nameof(Font));
		Font fontclass = new Font()
		{
			Assembly = Assembly,
			Embeded = true,
			Name = Font,
			HasTopFont = true
		};
		AllFonts.Add(fontclass);
		string Base = "GraphicsManager.Resources.Fonts.";
		if (Assembly is not null) Base = string.Empty;
		byte[] f = (Assembly is null
			? Tools.GetResourceBytes(Base + Font)
			: Tools.GetResourceBytes(Assembly!, $"{Base}{Font}"));
		byte[] hash = SHA256.HashData(f);
		if (AllMemoryFonts.ContainsKey(hash)) AllMemoryFonts.Add(hash, new(lib, f, 0));
		fontclass._Faces.Insert(0, AllMemoryFonts[hash]);
		return fontclass;
	}

	private static Font? Cache = null;
	public static Font MakeFontFromSystem(uint PixelHeight = 20)
	{
		if (Cache is null)
		{
			Cache = new() { PixelHeight = PixelHeight };
			AllFonts.Add(Cache);
		}
		return Cache;
	}
	public static Font MakeFontFromFile(string Font)
	{
		_ = Font ?? throw new ArgumentNullException(nameof(Font));
		Font fontclass = new Font()
		{
			Assembly = null,
			Embeded = false,
			Name = Font,
			HasTopFont = true
		};
		AllFonts.Add(fontclass);
		fontclass._Faces.Insert(0, new Face(lib, Font, 0));
		return fontclass;
	}
	
	public void SetFontFile(string Font)
	{
		_ = Font ?? throw new ArgumentNullException(nameof(Font));
		if (HasTopFont) _Faces[0] = new Face(lib, Font, 0);
		else _Faces.Insert(0, new Face(lib, Font, 0));
		HasTopFont = true;
		this.Assembly = null;
		this.Embeded = false;
		this.Name = Font;
	}
	
	private bool HasTopFont = false;
	public uint PixelHeight { get; set; } = 20;
	public string Name { get; private set; } = default!;
	public bool Embeded { get; private set; }
	public Assembly? Assembly { get; private set; }
}
