﻿using GraphicsManager.Structs;
using OpenTK.Graphics.ES30;
using OpenTK.Mathematics;
using OpenTK.Windowing.Desktop;
using SharpFont;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.PixelFormats;
using SixLabors.ImageSharp.Processing;
using GL = OpenTK.Graphics.OpenGL4.GL;
using Image = SixLabors.ImageSharp.Image;
using PixelFormat = OpenTK.Graphics.OpenGL4.PixelFormat;
using PixelInternalFormat = OpenTK.Graphics.OpenGL4.PixelInternalFormat;
using PixelStoreParameter = OpenTK.Graphics.OpenGL4.PixelStoreParameter;
using PixelType = OpenTK.Graphics.OpenGL4.PixelType;
using TextureMagFilter = OpenTK.Graphics.OpenGL4.TextureMagFilter;
using TextureMinFilter = OpenTK.Graphics.OpenGL4.TextureMinFilter;
using TextureParameterName = OpenTK.Graphics.OpenGL4.TextureParameterName;
using TextureTarget = OpenTK.Graphics.OpenGL4.TextureTarget;
using TextureUnit = OpenTK.Graphics.OpenGL4.TextureUnit;
using TextureWrapMode = OpenTK.Graphics.OpenGL4.TextureWrapMode;
using VertexAttribPointerType = OpenTK.Graphics.OpenGL4.VertexAttribPointerType;

namespace GraphicsManager.Objects.Core;

public class Texture
{
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> TextureShader = new ();

	public int handel;
	public Texture(byte[] File)
	{
		Image<Rgba32> image = Image.Load<Rgba32>(File);
		image.Mutate(x => x.Flip(FlipMode.Vertical));

		var pixels = new List<byte>(4 * image.Width * image.Height);

		for (int y = 0; y < image.Height; y++)
		{
			var row = image.GetPixelRowSpan(y);

			for (int x = 0; x < image.Width; x++)
			{
				pixels.Add(row[x].R);
				pixels.Add(row[x].G);
				pixels.Add(row[x].B);
				pixels.Add(row[x].A);
			}
		}

		handel = GL.GenTexture();
		GL.BindTexture(TextureTarget.Texture2D, handel);
		GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, image.Width, image.Height, 0, PixelFormat.Rgba, PixelType.UnsignedByte, pixels.ToArray());
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
	}

	private Texture()
	{
	}

	internal static Character GetChar(Font l, char charter)
	{
		int last = 0;
		for (int i = 0; i < l.Fonts.Count; i++)
		{
			try
			{
				try
				{
					l.Fonts[i].SetPixelSizes(0, l.PixelHeight);
					last = i;
				}
				catch (Exception e)
				{
					continue;
				}
				l.Fonts[i].SelectCharmap(Encoding.Unicode);
				ushort temp = ((ushort)charter);
				if (l.Fonts[i].GetCharIndex(temp) == 0) continue;
				l.Fonts[i].LoadChar(temp, LoadFlags.Render, LoadTarget.Normal);
				GlyphSlot glyph = l.Fonts[i].Glyph;
				FTBitmap bitmap = glyph.Bitmap;
				
				return new()
				{
					Size = new Vector2(bitmap.Width, bitmap.Rows),
					Bearing = new Vector2(glyph.BitmapLeft, glyph.BitmapTop),
					Advance = glyph.Advance.X.Value,
				};
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
				continue;
			}
		}
		for (int i = 0; i < Font._SystemPre.Count; i++)
		{
			try
			{
				Face tmp = new(Font.lib, Font._SystemPre[i], 0);
				try
				{
					tmp.SetPixelSizes(0, l.PixelHeight);
				}
				catch (Exception e)
				{
					tmp.Dispose();
					continue;
				}
				tmp.SelectCharmap(Encoding.Unicode);
				ushort temp2 = ((ushort)charter);
				if (tmp.GetCharIndex(temp2) == 0)
				{
					tmp.Dispose(); 
					continue;
				}
				tmp.LoadChar(temp2, LoadFlags.Render, LoadTarget.Normal);
				GlyphSlot glyph2 = tmp.Glyph;
				FTBitmap bitmap2 = glyph2.Bitmap;
				l.AddSystemFontFace(Font._SystemPre[i]);
				return new()
				{
					Size = new Vector2(bitmap2.Width, bitmap2.Rows),
					Bearing = new Vector2(glyph2.BitmapLeft, glyph2.BitmapTop),
					Advance = glyph2.Advance.X.Value,
				};
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
			}
		}
		l.Fonts[last].SetPixelSizes(0, l.PixelHeight);
		
		l.Fonts[last].SelectCharmap(Encoding.Unicode);
		ushort temp22 = ((ushort)charter);
		l.Fonts[last].LoadChar(temp22, LoadFlags.Render, LoadTarget.Normal);
		GlyphSlot glyph22 = l.Fonts[0].Glyph;
		FTBitmap bitmap22 = glyph22.Bitmap;

		return new()
		{
			Size = new Vector2(bitmap22.Width, bitmap22.Rows),
			Bearing = new Vector2(glyph22.BitmapLeft, glyph22.BitmapTop),
			Advance = (int)glyph22.Advance.X.Value,
		};

	}

	internal static Texture TextureForChar(IGLFWGraphicsContext con, Font l, char charter)
	{
		Texture? t = null;
		int last = 0;
		for (int i = 0; i < l.Fonts.Count; i++)
		{
			try
			{
				if (!Label._characters[con].ContainsKey(l)) Label._characters[con].Add(l, new Dictionary<uint, Character>());
				if (Label._characters[con][l].ContainsKey(charter)) return Label._characters[con][l][(ushort)charter].Texture;
				try
				{
					l.Fonts[i].SetPixelSizes(0, l.PixelHeight);
					last = i;
				}
				catch (Exception e)
				{
					continue;
				}
				

				GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);

				GL.ActiveTexture(TextureUnit.Texture0);
				l.Fonts[i].SelectCharmap(Encoding.Unicode);
				ushort temp = ((ushort)charter);
				if (l.Fonts[i].GetCharIndex(temp) == 0) continue;
				l.Fonts[i].LoadChar(temp, LoadFlags.Render, LoadTarget.Normal);
				GlyphSlot glyph = l.Fonts[i].Glyph;
				FTBitmap bitmap = glyph.Bitmap;
				t = new();
				t.handel = GL.GenTexture();
				GL.BindTexture(TextureTarget.Texture2D, t.handel);
				GL.TexImage2D(TextureTarget.Texture2D, 0,
					PixelInternalFormat.R8, bitmap.Width, bitmap.Rows, 0,
					PixelFormat.Red, PixelType.UnsignedByte, bitmap.Buffer);


				Character cha = new()
				{
					Size = new Vector2(bitmap.Width, bitmap.Rows),
					Bearing = new Vector2(glyph.BitmapLeft, glyph.BitmapTop),
					Advance = (int)glyph.Advance.X.Value,
					Texture = t,
				};

				Label._characters[con][l].Add(temp, cha);
				return t;
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
				continue;
			}
		}

		if (t is null)
		{
			for (int i = 0; i < Font._SystemPre.Count; i++)
			{
				try
				{
					Face tmp = new(Font.lib, Font._SystemPre[i], 0);
					try
					{
						tmp.SetPixelSizes(0, l.PixelHeight);
					}
					catch (Exception e)
					{
						tmp.Dispose();
						continue;
					}
					GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);

					GL.ActiveTexture(TextureUnit.Texture0);
					tmp.SelectCharmap(Encoding.Unicode);
					ushort temp2 = ((ushort)charter);
					if (tmp.GetCharIndex(temp2) == 0)
					{
						tmp.Dispose(); 
						continue;
					}
					tmp.LoadChar(temp2, LoadFlags.Render, LoadTarget.Normal);
					GlyphSlot glyph2 = tmp.Glyph;
					FTBitmap bitmap2 = glyph2.Bitmap;
					l.AddSystemFontFace(Font._SystemPre[i]);
					t = new();
					t.handel = GL.GenTexture();
					GL.BindTexture(TextureTarget.Texture2D, t.handel);
					GL.TexImage2D(TextureTarget.Texture2D, 0,
						PixelInternalFormat.R8, bitmap2.Width, bitmap2.Rows, 0,
						PixelFormat.Red, PixelType.UnsignedByte, bitmap2.Buffer);
					Character cha2 = new()
					{
						Size = new Vector2(bitmap2.Width, bitmap2.Rows),
						Bearing = new Vector2(glyph2.BitmapLeft, glyph2.BitmapTop),
						Advance = (int)glyph2.Advance.X.Value,
						Texture = t,
					};

					Label._characters[con][l].Add(temp2, cha2);
					return t;
				}
				catch (Exception ex)
				{
					Console.WriteLine(ex);
				}
			}
			if (!Label._characters[con].ContainsKey(l)) Label._characters[con].Add(l, new Dictionary<uint, Character>());
			if (Label._characters[con][l].ContainsKey(charter)) return Label._characters[con][l][(ushort)charter].Texture;
			l.Fonts[last].SetPixelSizes(0, l.PixelHeight);
				

			GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);

			GL.ActiveTexture(TextureUnit.Texture0);
			l.Fonts[last].SelectCharmap(Encoding.Unicode);
			ushort temp = ((ushort)charter);
			l.Fonts[last].LoadChar(temp, LoadFlags.Render, LoadTarget.Normal);
			GlyphSlot glyph = l.Fonts[0].Glyph;
			FTBitmap bitmap = glyph.Bitmap;
			if (t is null) t = new();
			else return t;
			t.handel = GL.GenTexture();
			GL.BindTexture(TextureTarget.Texture2D, t.handel);
			GL.TexImage2D(TextureTarget.Texture2D, 0,
				PixelInternalFormat.R8, bitmap.Width, bitmap.Rows, 0,
				PixelFormat.Red, PixelType.UnsignedByte, bitmap.Buffer);


			Character cha = new()
			{
				Size = new Vector2(bitmap.Width, bitmap.Rows),
				Bearing = new Vector2(glyph.BitmapLeft, glyph.BitmapTop),
				Advance = (int)glyph.Advance.X.Value,
				Texture = t,
			};

			Label._characters[con][l].Add(temp, cha);
		}

		return t;
	}

	public void LoadText()
	{
		GL.TextureParameter(handel, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
		GL.TextureParameter(handel, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
		GL.TextureParameter(handel, TextureParameterName.TextureWrapS, (int)TextureWrapMode.ClampToEdge);
		GL.TextureParameter(handel, TextureParameterName.TextureWrapT, (int)TextureWrapMode.ClampToEdge);
	}

	public void Load(int loc)
	{
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.Repeat);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.Repeat);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Nearest);
		GL.EnableVertexAttribArray(loc);
		GL.VertexAttribPointer(loc, 2, VertexAttribPointerType.Float, false, 5 * sizeof(float), 3 * sizeof(float));
	}

	public void Use(TextureUnit unit = TextureUnit.Texture0)
	{
		GL.ActiveTexture(unit);
		GL.BindTexture(TextureTarget.Texture2D, handel);
	}
}
